<%
    MAX_N = int(max_n)
%>\
// randomized protocol for signing contracts Even, Goldreich and Lempel

dtmc

// we now let B to makes his/her choices based on what he/she knows 
// to do this I have added non-determinism to the previous version
// and changed the modules so that only "B's view" is visible
// then reveal the values when B thinks he has an advantage

// to model the non-deterministic behaviour of corrupt party (party B)
// we have a set of possible initial states corresponding to what messages
// he/she tries to over hear when sending - we could do this with nondeterminism 
// but it will just make the model less structured and B has to make the choices 
// at the start anyway since B's view at this point should tell him nothing
// (we use the new construct init...endinit to specify the set of initial states)

// note that certain variables that belong to a party appear in the other party's module
// as this leads to a more structured model - without this PRISM runs out of memory

// note we have included the case when B stops if he/she thinks that the protocol has reached 
// a state where he/she has an advantage

// currently, this model only works for N up to ${MAX_N}

const int N; // number of pairs of secrets the party sends
const int L; // number of bits in each secret

module counter
  b : [1..L]; // counter for current bit to be send (used in phases 2 and 3)
  n : [0..max(N-1,1)]; // counter as parties send N messages in a row
  phase : [1..4]; // phase of the protocol
  party : [1..2]; // which party moves
  // 1 first phase of the protocol (sending messages of the form OT(.,.,.,.)
  // 2 and 3 - second phase of the protocol (sending secretes 1..n and n+1..2n respectively)
  // 4 finished the protocol

  // FIRST PHASE
  [receiveB] phase=1 & party=1 -> (party'=2); // first A sends a message then B does
  [receiveA] phase=1 & party=2 & n<N-1 -> (party'=1) & (n'=n+1); // after B sends a message we move onto the next message
  [receiveA] phase=1 & party=2 & n=N-1 -> (party'=1) & (phase'=2) & (n'=0); // B has sent his final message - move to next phase

  // SECOND AND THIRD PHASES
  // A sends bit b of secret 1
  [receiveB] ((phase)>=(2)&(phase)<=(3)) & party=1 & n=0   -> (party'=2);
  // B sends bit b of secrets 1..N or N+1..2N
  [receiveA] ((phase)>=(2)&(phase)<=(3)) & party=2 & n<N-1 -> (n'=n+1);
  [receiveA] ((phase)>=(2)&(phase)<=(3)) & party=2 & n=N-1 -> (party'=1) & (n'=1);
  // A sends bit b of secrets 2..N or N+2..2N
  [receiveB] ((phase)>=(2)&(phase)<=(3)) & party=1 & n<N-1 & n>0 -> (n'=n+1);
  [receiveB] ((phase)>=(2)&(phase)<=(3)) & party=1 & n=N-1 & b<L -> (party'=1) & (n'=0) & (b'=b+1); // finished for party B move to next bit
  [receiveB] phase=2 & party=1 & n=N-1   & b=L -> (phase'=3) & (party'=1) & (n'=0) & (b'=1); // finished for party B move to next phase
  [receiveB] phase=3 & party=1 & n=N-1   & b=L -> (phase'=4); // finished protocol (reveal values)

  // FINISHED
  [] phase=4 -> (phase'=4); // loop
endmodule

// party A
module partyA
  
  // bi the number of bits of B's ith secret A knows 
  // (keep pairs of secrets together to give a more structured model)
% for i in xrange(0, MAX_N):
  b${'{:<{}}'.format(i,len(str(MAX_N-1)))} : [0..L]; b${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))} : [0..L];
% endfor
  
  
  // first step (get either secret i or (N-1)+i with equal probability)
% for i in xrange(0, MAX_N):
  [receiveA] phase=1 & n=${'{:<{}}'.format(i,len(str(MAX_N-1)))} -> 0.5 : (b${'{:<{}}'.format(i,len(str(MAX_N-1)))}'=L)  + 0.5 : (b${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))}'=L);
% endfor
  // second step (secrets 0,...,N-1)
% for i in xrange(0, MAX_N):
  [receiveA] phase=2 & n=${'{:<{}}'.format(i,len(str(MAX_N-1)))} -> (b${'{:<{}}'.format(i,len(str(MAX_N-1)))}'=min(b${'{:<{}}'.format(i,len(str(MAX_N-1)))}+1,L));
% endfor
  // second step (secrets N,...,2N-1)
% for i in xrange(0, MAX_N):
  [receiveA] phase=3 & n=${'{:<{}}'.format(i,len(str(MAX_N-1)))} -> (b${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))}'=min(b${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))}+1,L));
% endfor
endmodule

// construct module for party B through renaming
module partyB=partyA[
% for i in xrange(0, 2*MAX_N):
                     b${'{:<{}}'.format(i,len(str(2*MAX_N-1)))} = a${'{:<{}}'.format(i,len(str(2*MAX_N-1)))},
% endfor
                     receiveA = receiveB] 
endmodule

// formulae
formula B_knows = ( \
% for i in xrange(0, MAX_N):
${'' if loop.first else '                  | '}(a${'{:<{}}'.format(i,len(str(2*MAX_N-1)))} = L & a${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))} = L)${');' if loop.last else ''}
% endfor

formula A_knows = ( \
% for i in xrange(0, MAX_N):
${'' if loop.first else '                  | '}(b${'{:<{}}'.format(i,len(str(2*MAX_N-1)))} = L & b${'{:<{}}'.format(MAX_N+i,len(str(2*MAX_N-1)))} = L)${');' if loop.last else ''}
% endfor

// labels
label "B_knows" = B_knows;
label "A_knows" = A_knows;

// reward structures

// messages from B that A needs to know a pair
rewards "messages_A"
//  [receiveA] !A_knows : 1;
  !A_knows & party=2 : 1;
endrewards

// messages from B that A needs to know a pair once B knows a pair
rewards "messages_A_needs"
//  [receiveA] B_knows & !A_knows : 1;
  B_knows & !A_knows & party=2: 1;
endrewards

// messages from A that B needs to know a pair
rewards "messages_B"
//  [receiveB] !B_knows : 1;
  !B_knows & party=1: 1;
endrewards

// messages from A that B needs to know a pair once A knows a pair
rewards "messages_B_needs"
//  [receiveB] A_knows & !B_knows : 1;
  A_knows & !B_knows & party=1 : 1;
endrewards

formula lower_bits_A_knows = (${' + '.join('b{:<{}}'.format(i, len(str(2*MAX_N-1))) for i in xrange(0,MAX_N))});
formula upper_bits_A_knows = (${' + '.join('b{:<{}}'.format(i, len(str(2*MAX_N-1))) for i in xrange(MAX_N,2*MAX_N))});
formula lower_bits_B_knows = (${' + '.join('a{:<{}}'.format(i, len(str(2*MAX_N-1))) for i in xrange(0,MAX_N))});
formula upper_bits_B_knows = (${' + '.join('a{:<{}}'.format(i, len(str(2*MAX_N-1))) for i in xrange(MAX_N,2*MAX_N))});